from scanpy import settings
from scanpy import logging as logg

from datetime import datetime
from time import time


def get_passed_time():
    now = time()
    elapsed = now - settings._previous_time
    settings._previous_time = now
    return elapsed


def get_date_string():
    return datetime.now().strftime("%Y-%m-%d %H:%M")


def print_version_and_date():
    from . import __version__
    logg._write_log('Running scvelo', __version__, 'on {}.'.format(get_date_string()))



from scanpy.api.pl import scatter
from sklearn.neighbors import NearestNeighbors
from scipy.stats import norm as normal
import numpy as np
import matplotlib.pyplot as pl


def compute_velocity_on_grid(X_emb, V_emb, density=1, smooth=0.5, n_neighbors=None, min_mass=.5):
    """Computes the velocities for the grid points on the embedding

    Arguments
    ---------
    X_emb: np.ndarray
        embedding coordinates

    V_emb: np.ndarray
        embedded single cell velocity (obtained with 'velocity_embedding')

    density: float, default=1

    smooth: float, default=.5

    n_neighbors: int

    min_mass: float, default=.5

    Returns
    -------
    grid_coord: np.array
        grid point coordinates
    grid_velocity: np.array
        velocity for each grid point in the embedding
    """
    # prepare grid
    n_obs, n_dim = X_emb.shape
    steps = (int(np.sqrt(n_obs) * density), int(np.sqrt(n_obs) * density))

    grs = []
    for dim_i in range(n_dim):
        m, M = np.min(X_emb[:, dim_i]), np.max(X_emb[:, dim_i])
        m = m - .01 * np.abs(M - m)
        M = M + .01 * np.abs(M - m)
        gr = np.linspace(m, M, steps[dim_i])
        grs.append(gr)

    meshes_tuple = np.meshgrid(*grs)
    grid_coord = np.vstack([i.flat for i in meshes_tuple]).T

    # estimate grid velocities
    if n_neighbors is None: n_neighbors = int(n_obs/50)
    nn = NearestNeighbors(n_neighbors=n_neighbors, n_jobs=-1)
    nn.fit(X_emb)
    dists, neighs = nn.kneighbors(grid_coord)

    std = np.mean([(g[1] - g[0]) for g in grs])
    weight = normal.pdf(loc=0, scale=smooth * std, x=dists)
    p_mass = weight.sum(1)

    grid_velocity = (V_emb[neighs] * weight[:, :, None]).sum(1) / np.maximum(1, p_mass)[:, None]
    grid_coord, grid_velocity = grid_coord[p_mass > min_mass], grid_velocity[p_mass > min_mass]

    return grid_coord, grid_velocity


def velocity_embedding_grid(adata, basis='umap', vbasis='velocity', density=1, color=None,
                            min_mass=.5, smooth=.5, n_neighbors=None, principal_curve=False,
                            use_raw=True, sort_order=True, alpha=.2, groups=None, components=None, projection='2d',
                            legend_loc='right margin', legend_fontsize=None, legend_fontweight=None,
                            color_map=None, palette=None, frameon=False, right_margin=None, left_margin=None,
                            size=None, title=None, show=True, save=None, ax=None, **kwargs):
    """Scatter plot with grid velocities along `.obs` or `.var` axes.
    Color the plot using annotations of observations (`.obs`), variables (`.var`) or expression of genes (`.var_names`).

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    basis: `str` (default: `'umap'`)
        Key for embedding coordinates.
    vbasis: `str` (default: `'velocity'`)
        Key for velocity embedding coordinates.
    color : `str` or `None` (default: `None`)
        Key for annotations of observations/cells or variables/genes.

    Returns
    -------
        `matplotlib.Axis` if `show==False`
    """
    if ax is None: ax = pl.figure(None, (14, 10), dpi=120).gca()

    scatter(adata, color=color, use_raw=use_raw, sort_order=sort_order, alpha=alpha, basis=basis,
            groups=groups, components=components, projection=projection, legend_loc=legend_loc,
            legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight, color_map=color_map,
            palette=palette, frameon=frameon, right_margin=right_margin, left_margin=left_margin,
            size=size, title=title, show=False, save=save, ax=ax)

    vbasis += '_' + basis
    if vbasis not in adata.obsm_keys():
        raise ValueError(
            'You need to run `tl.velocity_embedding` first to compute embedded velocity vectors.')

    X_emb = adata.obsm['X_' + basis][:, :2]
    V_emb = adata.obsm[vbasis]

    _kwargs = {"scale": .5, "width": .001, "color": 'black', "edgecolors": 'k', "headwidth": 4.5, "headlength": 5,
                   "headaxislength": 3, "linewidth": .2}
    _kwargs.update(kwargs)

    X, V = compute_velocity_on_grid(X_emb, V_emb, density, smooth, n_neighbors, min_mass)
    pl.quiver(X[:, 0], X[:, 1], V[:, 0], V[:, 1], angles='xy', scale_units='xy', **_kwargs)

    if principal_curve:
        curve = adata.uns['principal_curve']['projections']
        pl.plot(curve[:, 0], curve[:, 1], c="w", lw=6, zorder=1000000)
        pl.plot(curve[:, 0], curve[:, 1], c="k", lw=3, zorder=2000000)

    if show: pl.show()
    else: return ax


import numpy as np
import matplotlib.pyplot as pl
from matplotlib.colors import rgb2hex
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


def get_colors(adata, basis):
    cluster_ix = adata.obs[basis].cat.codes
    if basis+'_colors' in adata.uns.keys():
        clusters_colors = adata.uns[basis + '_colors']
    else:
        colormap = np.vstack((pl.cm.tab20b(np.linspace(0., 1, 20))[::2], pl.cm.tab20c(np.linspace(0, 1, 20))[1::2],
                              pl.cm.tab20b(np.linspace(0., 1, 20))[1::2], pl.cm.tab20c(np.linspace(0, 1, 20))[0::2]))
        clusters_colors = [rgb2hex(c) for c in colormap[:len(adata.obs[basis])]]
    return np.array([clusters_colors[cluster_ix[i]] for i in range(adata.X.shape[0])])


def interpret_colorkey(adata, c=None, layer=None, perc=None):
    if perc is None: perc = [2, 98]
    if isinstance(c, str):
        if c in adata.obs.keys():  # color by observation key
            if adata.obs[c].dtype.name == 'category':
                c = get_colors(adata, basis=c)
            else:
                c = adata.obs[c]
                lb, ub = np.percentile(c, perc)
                c = np.clip(c, lb, ub)
        elif c in adata.var_names:  # color by var in specific layer
            c = adata[:, c].layers[layer] if layer in adata.layers.keys() else adata[:, c].X
            lb, ub = np.percentile(c, perc)
            c = np.clip(c, lb, ub)
    elif c is None:  # color by cluster or louvain or grey if no color is specified
        c = get_colors(adata, 'clusters') if 'clusters' in adata.obs.keys() \
            else get_colors(adata, 'louvain') if 'louvain' in adata.obs.keys() else 'grey'
    elif isinstance(c, np.ndarray):  # continuous coloring
        lb, ub = np.percentile(c, perc)
        c = np.clip(c, lb, ub)
    else:
        raise ValueError('color key is invalid! pass valid observation annotation or a gene name')
    return c


def scatter(adata, x=None, y=None, basis='umap', layer=None, color=None, xlabel=None, ylabel=None, color_map=None,
            size=None, alpha=1, fontsize=None, frameon=False, title=None, show=True, colorbar=False, save=None, ax=None, **kwargs):
    """Scatter plot along observations or variables axes.
    Color the plot using annotations of observations (`.obs`), variables (`.var`) or expression of genes (`.var_names`).

    Arguments
    ---------
    adata: `AnnData`
        Annotated data matrix.

    basis: `str` (default='tsne')
        plots embedding obsm['X_' + basis]

    x: `str`, `np.ndarray` or `None` (default: `None`)
        x coordinate

    y: `str`, `np.ndarray` or `None` (default: `None`)
        y coordinate

    color : `str` or `None` (default: `None`)
        Key for annotations of observations/cells or variables/genes

    Returns
    -------
        If `show==False` a `matplotlib.Axis`
    """
    if (x is None) | (y is None):
            X_emb = adata.obsm['X_' + basis][:, :2]
            x, y = X_emb[:, 0], X_emb[:, 1]

    pl.scatter(x, y, c=interpret_colorkey(adata, color, layer), cmap=color_map, s=size, alpha=alpha, **kwargs)

    if isinstance(title, str):
        pl.title(title, fontsize=fontsize)

    if isinstance(xlabel, str) or isinstance(ylabel, str):
        pl.xlabel(xlabel, fontsize=fontsize)
        pl.ylabel(ylabel, fontsize=fontsize)

    if not frameon: pl.axis('off')

    if isinstance(save, str):
        pl.savefig(save)

    if ax is not None:
        ax.xaxis.set_major_locator(MaxNLocator(nbins=3))
        ax.yaxis.set_major_locator(MaxNLocator(nbins=3))
        labelsize = int(fontsize * .75) if fontsize is not None else None
        ax.tick_params(axis='both', which='major', labelsize=labelsize)

    if colorbar and (ax is not None):
        cb = pl.colorbar(orientation='vertical', cax= inset_axes(ax, width="2%", height="30%", loc=4, borderpad=0))
        cb.locator = (MaxNLocator(nbins=3))
        cb.update_ticks()

    if show: pl.show()
    else: return ax


def phase(adata, var=None, x=None, y=None, color='louvain', fits='all', xlabel='spliced', ylabel='unspliced',
          fontsize=None, show=True, ax=None, **kwargs):
    if isinstance(var, str) and (var in adata.var_names):
        if (x is None) or (y is None):
            ix = np.where(adata.var_names == var)[0][0]
            x, y = adata.layers['Ms'][:, ix], adata.layers['Mu'][:, ix]
    else:
        ValueError('var not found in adata.var_names.')

    ax = scatter(adata, x=x, y=y, color=color, frameon=True, title=var, xlabel=xlabel, ylabel=ylabel, ax=ax, **kwargs)

    xnew = np.linspace(0, x.max() * 1.02)
    fits = adata.layers.keys() if fits == 'all' else fits
    fits = [fit for fit in fits if 'velocity' in fit]
    for fit in fits:
        linestyle = '--' if 'stochastic' in fit else '-'
        pl.plot(xnew, adata.var[fit+'_gamma'][ix] / adata.var[fit+'_beta'][ix] * xnew
                + adata.var[fit+'_offset'][ix] / adata.var[fit+'_beta'][ix], c='k', linestyle=linestyle)

    if show: pl.show()
    else: return ax


import numpy as np


def prod_sum_obs(A, B):
    """dot product and sum over axis 0 (obs) equivalent to np.sum(A * B, 0)
    """
    return np.einsum('ij, ij -> j', A, B)


def prod_sum_var(A, B):
    """dot product and sum over axis 1 (var) equivalent to np.sum(A * B, 1)
    """
    return np.einsum('ij, ij -> i', A, B)


def norm(A):
    """computes the L2-norm along axis 1 (e.g. genes or embedding dimensions) equivalent to np.linalg.norm(A, axis=1)
    """
    return np.sqrt(np.einsum('ij, ij -> i', A, A))



from .utils import *
from scanpy.api.pl import scatter
from matplotlib.colors import is_color_like


def velocity_embedding(adata, basis='umap', vbasis='velocity', layer=None, density=1, color=None,
                       use_raw=True, sort_order=True, alpha=.2, groups=None, components=None, projection='2d',
                       legend_loc='right margin', legend_fontsize=None, legend_fontweight=None,
                       color_map=None, palette=None, frameon=False, right_margin=None, left_margin=None,
                       size=None, title=None, show=True, save=None, ax=None, **kwargs):
    """Scatter plot with velocities along `.obs` or `.var` axes.
    Color the plot using annotations of observations (`.obs`), variables (`.var`) or expression of genes (`.var_names`).

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    basis: `str` (default: `'umap'`)
        Key for embedding coordinates.
    vbasis: `str` (default: `'velocity'`)
        Key for velocity embedding coordinates.
    color : `str` or `None` (default: `None`)
        Key for annotations of observations/cells or variables/genes.

    Returns
    -------
        `matplotlib.Axis` if `show==False`
    """
    if ax is None: ax = pl.figure(None, (14, 10), dpi=160).gca()

    scatter(adata, color=color, use_raw=use_raw, sort_order=sort_order, alpha=alpha, basis=basis,
            groups=groups, components=components, projection=projection, legend_loc=legend_loc,
            legend_fontsize=legend_fontsize, legend_fontweight=legend_fontweight, color_map=color_map,
            palette=palette, frameon=frameon, right_margin=right_margin, left_margin=left_margin,
            size=size, title=title, show=False, save=save, ax=ax)

    vbasis += '_' + basis
    if vbasis not in adata.obsm_keys():
        raise ValueError(
            'You need to run `tl.velocity_embedding` first to compute embedded velocity vectors.')

    X_emb = adata.obsm['X_' + basis][:, :2]
    V_emb = adata.obsm[vbasis]

    _kwargs = {"scale": 1, "width": .0005, "edgecolors": 'k', "headwidth": 9, "headlength": 10, "headaxislength": 6, "linewidth": .25}
    _kwargs.update(kwargs)

    ix_choice = np.random.choice(adata.n_obs, size=int(density * adata.n_obs), replace=False)

    X, V, C = X_emb[ix_choice], V_emb[ix_choice], interpret_colorkey(adata, color, layer)[ix_choice]

    if is_color_like(C[0]):
        pl.quiver(X[:, 0], X[:, 1], V[:, 0], V[:, 1], color=C, angles='xy', scale_units='xy', cmap=color_map, **_kwargs)
    else:
        pl.quiver(X[:, 0], X[:, 1], V[:, 0], V[:, 1], C, angles='xy', scale_units='xy', cmap=color_map, **_kwargs)

    if show: pl.show()
    else: return ax


from scipy.sparse import csr_matrix
import numpy as np
from ..logging import *


def moments(adata, renormalize=False, mode='connectivities', copy=False):
    """Computes first order moments for velocity estimation.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    renormalize: `bool` (default: `True`)
        Renormalize the moments by total counts per cell to its median.
    mode: `'connectivities'` or `'distances'`  (default: `'connectivities'`)
        Distance metric to use for moment computation.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to adata.

    Returns
    -------
    Returns or updates `adata` with the attributes
    Ms: `.layers`
        dense matrix with first order moments of spliced counts.
    Mu: `.layers`
        dense matrix with first order moments of unspliced counts.
    """
    if 'neighbors' not in adata.uns:
        raise ValueError('You need to run `pp.neighbors` first to compute a neighborhood graph.')

    logg.info('computing moments', r=True)

    if mode in adata.uns['neighbors']: connectivities = adata.uns['neighbors'][mode]
    else: raise ValueError('mode can only be  \'connectivities\' or \'distances\'')

    connectivities.setdiag(1)
    connectivities = connectivities.multiply(1. / connectivities.sum(1))

    connectivities += connectivities.dot(connectivities*.5)

    adata.layers['Ms'] = csr_matrix.dot(connectivities, csr_matrix(adata.layers['spliced'])).A
    adata.layers['Mu'] = csr_matrix.dot(connectivities, csr_matrix(adata.layers['unspliced'])).A

    if renormalize:
        adata.layers['Ms'] = adata.layers['Ms'] / adata.layers['Ms'].sum(1)[:, None] * np.median(adata.layers['Ms'].sum(1))
        adata.layers['Mu'] = adata.layers['Mu'] / adata.layers['Mu'].sum(1)[:, None] * np.median(adata.layers['Mu'].sum(1))

    logg.info('    finished', time=True, end=' ' if settings.verbosity > 2 else '\n')
    logg.hint(
        'added to `.layers`\n'
        '    \'Ms\', moments of spliced abundances\n'
        '    \'Mu\', moments of unspliced abundances')

    return adata if copy else None


def second_order_moments(adata):
    """Computes second order moments for stochastic velocity estimation.

    Arguments
    ---------
    adata: `AnnData`
        Annotated data matrix.

    Returns
    -------
    Mss: Second order moments for spliced abundances
    Mus: Second order moments for spliced with unspliced abundances
    """
    if 'neighbors' not in adata.uns:
        raise ValueError(
            'You need to run `pp.neighbors` first to compute a neighborhood graph.')

    connectivities = adata.uns['neighbors']['connectivities'] > 0
    connectivities = connectivities.multiply(1. / connectivities.sum(1))

    s, u = csr_matrix(adata.layers['spliced']), csr_matrix(adata.layers['unspliced'])
    Mss = csr_matrix.dot(connectivities, s.multiply(s)).A
    Mus = csr_matrix.dot(connectivities, s.multiply(u)).A

    return Mss, Mus


def second_order_moments_u(adata):
    """Computes second order moments for stochastic velocity estimation.

    Arguments
    ---------
    adata: `AnnData`
        Annotated data matrix.

    Returns
    -------
    Muu: Second order moments for unspliced abundances
    """
    if 'neighbors' not in adata.uns:
        raise ValueError(
            'You need to run `pp.neighbors` first to compute a neighborhood graph.')

    connectivities = adata.uns['neighbors']['connectivities'] > 0
    connectivities = connectivities.multiply(1. / connectivities.sum(1))

    u = csr_matrix(adata.layers['unspliced'])
    Muu = csr_matrix.dot(connectivities, u.multiply(u)).A

    return Muu


from .moments import moments


def recipe_velocity(adata, min_counts=10, n_top_genes=None, n_pcs=30, n_neighbors=30, log=True, copy=False):
    """Filtering, normalization, neighbors graph and moments.

    Expects non-logarithmized data. If using logarithmized data, pass `log=False`.

    The recipe runs the following steps

    .. code:: python

        sc.pp.filter_genes(adata, min_counts=10)
        sc.pp.normalize_per_cell(adata, key_n_counts='n_counts_all')
        sc.pp.filter_genes_dispersion(adata, n_top_genes=3000)
        sc.pp.normalize_per_cell(adata, layers='all)
        if log: sc.pp.log1p(adata)
        sc.pp.pca(adata, n_comps=30)
        sc.pp.neighbors(adata, n_neighbors=30, use_rep='X_pca')
        scv.pp.moments(adata)


    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    n_top_genes: `int` (default: 1000)
        Number of genes to keep.
    log: `bool` (default: `True`)
        Take logarithm.
    n_pcs: `int` (default: 30)
        Number of principal components to use.
    n_neighbors: `int` (default: 30)
        Number of neighbors to use.
    copy: `bool` (default: `False`)
        Return a copy of `adata` instead of updating it.

    Returns
    -------
    Returns or updates `adata` depending on `copy`.
    """
    from scanpy.api.pp import filter_genes, filter_genes_dispersion, normalize_per_cell, log1p, pca, neighbors

    filter_genes(adata, min_counts=min_counts)
    normalize_per_cell(adata, key_n_counts='n_counts_all')

    filter_genes_dispersion(adata, n_top_genes=n_top_genes)
    normalize_per_cell(adata, layers='all')
    if log: log1p(adata)
    pca(adata, n_comps=n_pcs, svd_solver='arpack')
    neighbors(adata, n_neighbors=n_neighbors, use_rep='X_pca')

    moments(adata)
    return adata if copy else None



from scipy.spatial.distance import pdist, squareform
from scipy.sparse import csr_matrix
import numpy as np
import warnings


def transition_matrix(adata, vkey='velocity', basis=None, direction="forward", scale=10):
    """Computes transition probabilities by applying Gaussian kernel to cosine similarities x scale

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` (default: `'velocity'`)
        Name of velocity estimates to be used.
    basis: `str` or `None` (default: `None`)
        Restrict transition to embedding if specified
    direction: `'forward'` or `'backward'` (default: `'foward'`)
        Whether to use the transition matrix to push forward or to pull backward
    scale: `float` (default: 10)
        Scale parameter of gaussian kernel.

    Returns
    -------
    Returns sparse matrix with transition probabilities.
    """
    if vkey+'_graph' not in adata.uns:
        raise ValueError(
            'You need to run `tl.velocity_graph` first to compute cosine correlations.')

    T = np.expm1(adata.uns[vkey + '_graph'] * scale)

    dists = adata.uns['neighbors']['distances']
    T = .5 * T + .5 * (dists > 0).multiply(T)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        T = T.multiply(csr_matrix(1. / T.sum(1)))

    if basis in adata.obsm.keys():
        if direction == 'backward': T = T.T
        X_emb = adata.obsm['X_' + basis]
        dists_emb = (T > 0).multiply(squareform(pdist(X_emb)))
        kernel = np.expm1(-dists_emb**2 / dists_emb.data.mean()**2 / 2)
        T = T.multiply(kernel)
        T = T.multiply(csr_matrix(1. / T.sum(1)))

    return T
